#!/usr/bin/python
# converts a view specification file to C++ header files

import sys
import os
import string

view = None
views = {}

class Attribute:
	def __init__(self, name, type, comments):
		self.name = name
		self.type = type
		self.comments = comments

class View:
	def __init__(self, comments):
		self.comments = comments
		self.attributes = {}

if len(sys.argv) != 4:
	print "USAGE: %s <input file> <output file> <namespace>" % (sys.argv[0])
	sys.exit(1)

input = open(sys.argv[1], 'r')
output = open(sys.argv[2], 'w')
hdrname = string.replace(sys.argv[2], '.', '_')
hdrname = string.replace(hdrname, '/', '_')
hdrname = '__%s__' % (hdrname)
namespace = sys.argv[3]
comments = []

line = input.readline()
while line:
	words = string.split(line)
	if len(words) > 0:
		name = words[0]
		if name[-1] == ':':
			view = name[:-1]
			views[view] = View(comments)
			comments = []
		elif name[0] == '#':
			comments.append(string.join(words[1:]))
		else:
			attribute = name
			type = words[1]
			if view:
				views[view].attributes[attribute] = Attribute(attribute, type, comments)
			comments = []


	line = input.readline()

pfx = "m_a_"
content = """/* generated by mkview.py */

#ifndef %s
#define %s

#include "fex/datatool.h"

namespace %s
{
""" % (hdrname, hdrname, namespace)

def mk_setup(view):
	content = ""
	for attr in views[view].attributes.keys():
		type = views[view].attributes[attr].type
		content += '\n\t\t\t%s%s.setup(spScope, "%s");' % (pfx, attr, attr)
		content += '\n\t\t\tm_accessors.push_back(%s%s);' % (pfx, attr)
	return content

def mk_access(view):
	content = ""
	for attr in views[view].attributes.keys():
		type = views[view].attributes[attr].type
		for comment in views[view].attributes[attr].comments:
			content += '\n\t\t/// %s' % (comment)
		content += """
		%s &%s(void)
		{
			if(!m_record.isValid())
			{
				feX("invalid record");
			}
			return %s%s(m_record);
		}""" % (type, attr, pfx, attr)
		for comment in views[view].attributes[attr].comments:
			content += '\n\t\t/// %s' % (comment)
		content += """
		%s &%s(Record &a_record)
		{
			if(!a_record.isValid())
			{
				feX("invalid record");
			}
			return %s%s(a_record);
		}""" % (type, attr, pfx, attr)
	return content

def mk_members(view):
	content = ""
	for attr in views[view].attributes.keys():
		type = views[view].attributes[attr].type
		content += "\n\t\tAccessor<%s> %s%s;" % (type, pfx, attr)
	return content

for view in views.keys():
	for comment in views[view].comments:
		content += '\n/// %s' % (comment)
	content += """
class %sView : public View
{
	public:
		%sView(sp<Scope> spScope)
		{%s
		}%s
	private:%s
};
""" % (view, view, mk_setup(view), mk_access(view), mk_members(view))

content += """
} /* namespace */

#endif
"""

output.write(content)

input.close()
output.close()

