/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __maya_SurfaceAccessibleMaya_h__
#define __maya_SurfaceAccessibleMaya_h__

namespace fe
{

/**************************************************************************//**
    @brief Maya Surface Binding

	@ingroup maya
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleMaya:
	public SurfaceAccessibleBase
{
	public:
								SurfaceAccessibleMaya(void):
									m_pMeshArrayData(NULL),
									m_pFnMesh(NULL)							{}
virtual							~SurfaceAccessibleMaya(void)
								{
									if(m_pFnMesh)
									{
										delete m_pFnMesh;
									}
								}

								//* as SurfaceAccessibleI
virtual	void					bind(Instance a_instance)
								{	setMeshData(
											a_instance.cast<MDataHandle>()); }
virtual	BWORD					isBound(void)
								{	return (m_pMeshArrayData ||
											!m_meshData.asMesh().isNull()); }

const	std::vector<String>&	groupNames(SurfaceAccessibleI::Element
										a_element) const;

virtual
const	std::vector<SurfaceAccessibleI::Spec>&
								attributeSpecs(SurfaceAccessibleI::Element
										a_element) const;

virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										String a_name,
										SurfaceAccessibleI::Creation a_create)
								{
									sp<SurfaceAccessorMaya> spAccessor(
											new SurfaceAccessorMaya);
									spAccessor->setMaster(registry()->master());
									spAccessor->setMeshNode(m_meshNode);
									spAccessor->setMeshData(m_meshData);
									spAccessor->setMesh(m_pFnMesh);
									if(spAccessor->bind(a_element,a_name) ||
											a_create==
											SurfaceAccessibleI::e_createMissing)
									{
										spAccessor->setSurfaceAccessible(
												sp<SurfaceAccessibleI>(this));
										return spAccessor;
									}
									return sp<SurfaceAccessorI>(NULL);
								}

virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute,
										SurfaceAccessibleI::Creation a_create)
								{
									if(a_attribute==
											SurfaceAccessibleI::e_normal &&
											threading()==
											SurfaceAccessibleI::e_multiThread)
									{
										//* normal writing not threadsafe, ever
										return sp<SurfaceAccessorI>(NULL);
									}
									sp<SurfaceAccessorMaya> spAccessor(
											new SurfaceAccessorMaya);
									spAccessor->setMaster(registry()->master());
									spAccessor->setMeshNode(m_meshNode);
									spAccessor->setMeshData(m_meshData);
									spAccessor->setMesh(m_pFnMesh);
									if(spAccessor->bind(a_element,a_attribute)
											|| a_create==
											SurfaceAccessibleI::e_createMissing)
									{
										spAccessor->setSurfaceAccessible(
												sp<SurfaceAccessibleI>(this));
										return spAccessor;
									}
									return sp<SurfaceAccessorI>(NULL);
								}

virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										String a_name);

virtual	sp<SurfaceI>			surface(String a_group);

								//* TODO
/*
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group);
*/

virtual	void					append(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const SpatialTransform& a_rTransform);

virtual	void					instance(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const std::vector<SpatialTransform>&
										a_rTransformArray);

								//* Maya specific
		void					setMeshNode(const MObject a_meshNode)
								{
									m_meshNode=a_meshNode;
								}
		void					setMeshData(const MDataHandle a_meshData)
								{
									m_meshData=a_meshData;
									m_spSurfaceI=NULL;

									if(m_pMeshArrayData)
									{
										delete m_pMeshArrayData;
										m_pMeshArrayData=NULL;
									}

									if(m_pFnMesh)
									{
										delete m_pFnMesh;
									}

									//* not thread safe
									m_pFnMesh=new MFnMesh(m_meshData.asMesh());
								}
const	MDataHandle				meshData(void)
								{	return m_meshData; }

		void					setMeshArrayData(
										const MArrayDataHandle& a_meshArrayData)
								{
									m_pMeshArrayData=new MArrayDataHandle(
											a_meshArrayData);
									m_spSurfaceI=NULL;

									m_meshData=MDataHandle();

									if(m_pFnMesh)
									{
										delete m_pFnMesh;
									}

									m_pFnMesh=NULL;
								}
		MArrayDataHandle&		meshArrayData(void)
								{	return *m_pMeshArrayData; }

	protected:

virtual	void					reset(void);

	private:
		sp<SurfaceI>		m_spSurfaceI;

		MObject				m_meshNode;
		MDataHandle			m_meshData;
		MArrayDataHandle*	m_pMeshArrayData;
		MFnMesh*			m_pFnMesh;
};

} // namespace

#endif /* __maya_SurfaceAccessibleMaya_h__ */
