/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __opengl_DrawOpenGL_h__
#define __opengl_DrawOpenGL_h__

#define DGL_CYLINDER_GLU	FALSE	// replace default with glu version
#define DGL_SPHERE_GLU		TRUE	// replace default with glu version

namespace fe
{

/**************************************************************************//**
	@brief OpenGL-specific implementations for DrawI

	@ingroup opengl
*//***************************************************************************/
class FE_DL_EXPORT DrawOpenGL: public DrawCommon,
		public Initialize<DrawOpenGL>
{
	public:
					DrawOpenGL();
virtual				~DrawOpenGL();

		void		initialize(void);

virtual	BWORD		isDirect(void) const				{ return m_isDirect; }
virtual	void		flush(void);

virtual void		setDrawMode(sp<DrawMode> spMode);

virtual	void		setFontBase(U32 index)			{ m_fontbase=index; };

virtual void		drawPoints(const SpatialVector *vertex,
							const SpatialVector *normal,U32 vertices,
							BWORD multicolor,const Color *color);

virtual void		drawLines(const SpatialVector *vertex,
							const SpatialVector *normal,U32 vertices,
							StripMode strip,BWORD multicolor,
							const Color *color);

virtual	void		drawTriangles(const SpatialVector *vertex,
							const SpatialVector *normal,
							const Vector2f *texture,U32 vertices,
							StripMode strip,BWORD multicolor,
							const Color *color);

virtual void		drawRectangles(const SpatialVector *vertex,U32 vertices,
						BWORD multicolor,const Color *color);

virtual	void		drawAlignedText(const SpatialVector &location,
							const String text,const Color &color);

#if DGL_CYLINDER_GLU
virtual	void		drawCylinder(const SpatialTransform &transform,
							const SpatialVector *scale,
							Real baseScale,const Color &color,U32 slices)
					{	drawCylinderGLU(transform,scale,baseScale,
								color,slices); }
#endif

#if DGL_SPHERE_GLU
virtual	void		drawSphere(const SpatialTransform &transform,
							const SpatialVector *scale,const Color &color)
					{	drawSphereGLU(transform,scale,color); }
#endif

virtual void		drawTransformedTriangles(
						const SpatialTransform &transform,
						const SpatialVector *scale,
						const SpatialVector *vertex,
						const SpatialVector *normal,
						const Vector2f *texture,
						U32 vertices,StripMode strip,BWORD multicolor,
						const Color *color);

virtual	void		drawRaster(sp<ImageI> spImageI,
						const SpatialVector& location);

	protected:

virtual	void		createDefaultView(void);

	private:

		void		threadLock(void);
		void		threadUnlock(void);

		void		bindTexture(sp<ImageI> spImageI);

		void		drawCylinderGLU(const SpatialTransform &transform,
							const SpatialVector *scale,
							Real baseScale,const Color &color,U32 slices);

		void		drawSphereGLU(const SpatialTransform &transform,
							const SpatialVector *scale,const Color &color);

		void		drawPolys(U32 grain,const SpatialVector *vertex,
							const SpatialVector *normal,
							const Vector2f *texture,
							U32 vertices,StripMode strip,BWORD multicolor,
							const Color *color);

		void		setupLights(void);
		void		createDefaultTexture(void);
		void		useDefaultTexture(void);
		void		pushAntialiasPoints(void);
		void		pushAntialiasLines(void);
		void		pushBlending(void);
		void		checkForError(void);
		void		queryFormat(sp<ImageI> spImageI,GLenum& format,
							GLenum& type,GLint& components);

		void		setup(DrawMode::DrawStyle style,BWORD lit,BWORD dual,
							BWORD texture,BWORD frontcull,BWORD backcull,
							BWORD zBuffer);

		sp<ImageI>			m_currentTexture;
		U32					m_fontbase;
		BWORD				m_isDirect;
		BWORD				m_lockThreads;
		GLUquadric*			m_pQuadric;
		GLubyte				m_texture[64][64][3];
};

} // namespace

#endif /* __opengl_DrawOpenGL_h__ */
