/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SBO_DEBUG	FALSE
#define FE_SBO_VERBOSE	FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceBindOp::initialize(void)
{
	//* page Config
	catalog<String>("Threads","page")="Config";
	catalog<String>("AutoThread","page")="Config";
	catalog<String>("Paging","page")="Config";
	catalog<String>("StayAlive","page")="Config";

	catalog<String>("BindPrefix")="bind";
	catalog<String>("BindPrefix","label")="Bind Attribute Prefix";
	catalog<String>("BindPrefix","page")="Config";
	catalog<bool>("BindPrefix","joined")=true;
	catalog<String>("BindPrefix","hint")=
			"Prefix of attribute names added to output surface describing"
			" where each point on the input surface is attached to"
			" the associated driver.";

	catalog<String>("InputGroup")="";
	catalog<String>("InputGroup","label")="Input Group";
//	catalog<String>("InputGroup","suggest")="pointGroups";	(point or primitive)
	catalog<String>("InputGroup","page")="Config";
	catalog<String>("InputGroup","hint")=
			"Group in Driver Surface that limit which"
			" point or primitives are affected."
			"  If no group is specified, the entire input is affected.";

	catalog<String>("DriverGroup")="";
	catalog<String>("DriverGroup","label")="Driver Group";
	catalog<String>("DriverGroup","suggest")="primitiveGroups";
	catalog<String>("DriverGroup","page")="Config";
	catalog<String>("DriverGroup","hint")=
			"Primitive Group in Driver Surface that encompass the binding."
			"  Primitives outside this group will not be"
			" considered during binding."
			"  If no group is specified, the entire driver is made available.";

	//* page Bind
	catalog<bool>("PartitionDriver")=false;
	catalog<String>("PartitionDriver","label")="Partition Driver";
	catalog<String>("PartitionDriver","page")="Bind";
	catalog<bool>("PartitionDriver","joined")=true;
	catalog<String>("PartitionDriver","hint")=
			"Restrict bindings by matching attribute values.";

	catalog<String>("DriverPartitionAttr")="part";
	catalog<String>("DriverPartitionAttr","label")="Driver Attr";
	catalog<String>("DriverPartitionAttr","page")="Bind";
	catalog<String>("DriverPartitionAttr","enabler")="PartitionDriver";
	catalog<bool>("DriverPartitionAttr","joined")=true;
	catalog<String>("DriverPartitionAttr","hint")=
			"Primitive string attribute on driver to use for bind matching."
			"  Generally, these values are just simple names.";

	catalog<String>("InputPartitionAttr")="part";
	catalog<String>("InputPartitionAttr","label")="Input Attr";
	catalog<String>("InputPartitionAttr","page")="Bind";
	catalog<String>("InputPartitionAttr","enabler")="PartitionDriver";
	catalog<String>("InputPartitionAttr","hint")=
			"Point string attribute on input to use for bind matching."
			"  Each value can be a simple name or a full regex pattern."
			"  For example, use \"prefix.*\" for every name starting"
			" with particular value"
			" (the '.' is required to indicate 'any character' and"
			" the '*' means 'repeated without limit')"
			" or \"one|two|three\" to allow any one name from a list."
			"  Be careful with non-letter characters that may have"
			" a special meaning in regular expressions.";

	catalog<bool>("subdiv")=false;
	catalog<String>("subdiv","label")="Subdivide Driver";
	catalog<String>("subdiv","page")="Bind";
	catalog<bool>("subdiv","joined")=true;
	catalog<String>("subdiv","hint")="Apply subdivision to driver surfaces.";

	catalog<I32>("subdivDepth")=1;
	catalog<I32>("subdivDepth","min")=1;
	catalog<I32>("subdivDepth","high")=4;
	catalog<I32>("subdivDepth","max")=10;
	catalog<String>("subdivDepth","label")="Depth";
	catalog<String>("subdivDepth","page")="Bind";
	catalog<String>("subdivDepth","enabler")="subdiv";
	catalog<String>("subdivDepth","hint")=
			"Number of subdivision iteration to apply to driver surfaces.";

	catalog<bool>("LimitDistance")=false;
	catalog<String>("LimitDistance","label")="Limit Distance";
	catalog<String>("LimitDistance","page")="Bind";
	catalog<bool>("LimitDistance","joined")=true;
	catalog<String>("LimitDistance","hint")=
			"Dismiss potential bindings beyond a specific distance.";

	catalog<Real>("MaxDistance")=1.0;
	catalog<Real>("MaxDistance","high")=10.0;
	catalog<Real>("MaxDistance","max")=1e6;
	catalog<String>("MaxDistance","label")="To";
	catalog<String>("MaxDistance","page")="Bind";
	catalog<String>("MaxDistance","enabler")="LimitDistance";
	catalog<String>("MaxDistance","hint")=
			"Disallow bindings beyond this distance.";

	catalog<bool>("Bind Fragments")=false;
	catalog<String>("Bind Fragments","page")="Bind";
	catalog<bool>("Bind Fragments","joined")=true;
	catalog<String>("Bind Fragments","hint")=
			"Bind collections of points or primitives with"
			" matching fragment attribute";

	catalog<String>("FragmentAttr")="part";
	catalog<String>("FragmentAttr","label")="Attr";
	catalog<String>("FragmentAttr","page")="Bind";
	catalog<String>("FragmentAttr","enabler")="Bind Fragments";
	catalog<String>("FragmentAttr","hint")=
			"Name of string attribute describing membership"
			" in a collection of primitives";

	catalog<bool>("generateLocators")=false;
	catalog<String>("generateLocators","label")="Generate Locators";
	catalog<String>("generateLocators","page")="Bind";
	catalog<String>("generateLocators","enabler")="Bind Fragments";
	catalog<bool>("generateLocators","joined")=true;
	catalog<String>("generateLocators","hint")=
			"Add detail string attributes describing the location"
			" and orientation of each fragment.";

	catalog<String>("locatorPrefix")="loc";
	catalog<String>("locatorPrefix","label")="Prefix";
	catalog<String>("locatorPrefix","page")="Bind";
	catalog<String>("locatorPrefix","enabler")="Bind Fragments";
	catalog<String>("locatorPrefix","hint")=
			"Text at the beginning of every locator attribute.";

	catalog<bool>("Bind Primitives")=false;
	catalog<String>("Bind Primitives","page")="Bind";
	catalog<bool>("Bind Primitives","joined")=true;
	catalog<String>("Bind Primitives","hint")=
			"Affect whole primitives instead of individual points";

	catalog<bool>("Conform")=false;
	catalog<String>("Conform","label")="Conform Primitives";
	catalog<String>("Conform","page")="Bind";
	catalog<String>("Conform","hint")=
			"Force all points of each input primitive to bind to"
			" the same subset of driver primitives."
			"  This can be useful to coerce all points on"
			" each input curve to follow the same driver curve(s).";

	catalog<bool>("Conform","joined")=true;
	catalog<bool>("Scalable")=false;
	catalog<String>("Scalable","page")="Bind";

	catalog<I32>("Bindings")=1;
	catalog<I32>("Bindings","min")=1;
	catalog<I32>("Bindings","high")=100;
	catalog<I32>("Bindings","max")=1e3;
	catalog<String>("Bindings","page")="Bind";
	catalog<bool>("Bindings","joined")=true;
	catalog<String>("Bindings","hint")=
			"Maximum number of attachment contacts.";

	catalog<bool>("Covert")=false;
	catalog<String>("Covert","label")="Covert Bindings";
	catalog<String>("Covert","page")="Bind";
	catalog<String>("Covert","hint")=
			"Store bindings apart from the surface.";

	catalog<String>("PivotPoint")="First";
	catalog<String>("PivotPoint","label")="Pivot Point";
	catalog<String>("PivotPoint","page")="Bind";
	catalog<String>("PivotPoint","choice:0")="First";
	catalog<String>("PivotPoint","choice:1")="Center";
	catalog<String>("PivotPoint","choice:2")="Nearest";
	catalog<String>("PivotPoint","hint")=
			"Location used to determine attachment.";

	catalog< sp<Component> >("Input Surface");
	catalog<String>("Input Surface","label")="Reference Input Surface";

	catalog< sp<Component> >("Driver Surface");
	catalog<String>("Driver Surface","label")="Reference Driver Surface";

	m_spSubdivideOp=registry()->create("OperatorSurfaceI.SubdivideOp");
}

void SurfaceBindOp::bind(sp<SignalerI> a_spSignalerI,sp<Layout> a_spLayout)
{
	OperatorSurfaceCommon::bind(a_spSignalerI,a_spLayout);

	sp<OperatorSurfaceCommon> spCommon=m_spSubdivideOp;
	if(spCommon.isValid())
	{
		spCommon->bind(a_spSignalerI,a_spLayout);
	}
}

void SurfaceBindOp::handle(Record& a_rSignal)
{
#if FE_SBO_DEBUG
	feLog("SurfaceBindOp::handle\n");
#endif

	const String inputGroup=catalog<String>("InputGroup");
	m_partitionDriver=catalog<bool>("PartitionDriver");
	m_bindPrimitives=catalog<bool>("Bind Primitives");
	m_conform=!m_bindPrimitives && catalog<bool>("Conform");

	const String fragmentAttribute=catalog<String>("FragmentAttr");
	const String driverPartitionAttr=catalog<String>("DriverPartitionAttr");
	const BWORD covert=catalog<bool>("Covert");
	const BWORD subdiv=catalog<bool>("subdiv");
	const I32 subdivDepth=subdiv? catalog<I32>("subdivDepth"): 0;

	BWORD warnedConformTriangle=FALSE;

	if(m_bindPrimitives && catalog<bool>("Conform"))
	{
		catalog<String>("warning")+=
				"Conforming only supports point bindings;";
	}

	//* output surface is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	m_element=m_bindPrimitives? e_primitive: e_point;
	m_attribute=m_bindPrimitives? e_vertices: e_position;

	const String covertKey=name();	//* TODO path also?

	m_spBindingAccessible=m_spOutputAccessible;
	if(covert)
	{
		sp<SurfaceAccessibleCatalog> spSurfaceAccessibleCatalog=
				registry()->create("*.SurfaceAccessibleCatalog");
		if(spSurfaceAccessibleCatalog.isValid())
		{
			spSurfaceAccessibleCatalog->setCatalog(
					registry()->master()->catalog());
			spSurfaceAccessibleCatalog->setKey(covertKey);
			m_spBindingAccessible=spSurfaceAccessibleCatalog;
		}
	}

	const String driverGroup=catalog<String>("DriverGroup");

	sp<SurfaceAccessibleI> spDriverAccessible;
	if(!access(spDriverAccessible,"Driver Surface")) return;

	sp<SurfaceAccessorI> spDriverNormals;
	if(!access(spDriverNormals,"Driver Surface",e_point,e_normal) ||
			!spDriverNormals->count())
	{
		catalog<String>("warning")+=
				"Driver Surface has no normals (expect rotational problems);";
	}
	spDriverNormals=NULL;

	if(subdivDepth)
	{
		m_spDriver=subdivide(spDriverAccessible,driverGroup,
				subdivDepth,0,a_rSignal);
	}
	else
	{
		if(!access(m_spDriver,spDriverAccessible,driverGroup)) return;
	}

	if(m_partitionDriver)
	{
		m_spDriver->partitionWith(driverPartitionAttr);
	}

	if(m_spDriver->radius()<1e-6)
	{
		catalog<String>("error")=
				"Driver Surface appears empty (zero radius bounding sphere);";
		return;
	}

	m_spDriver->prepareForSearch();

	sp<SurfaceAccessorI> spOutputVertices;
	if((m_conform || m_bindPrimitives) && !access(spOutputVertices,
			m_spOutputAccessible,e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	const BWORD bindFragments=catalog<bool>("Bind Fragments");
	m_fragmented=(bindFragments && spOutputElements.isValid());

	I32 fragmentCount=1;
	if(m_fragmented)
	{
		spOutputElements->fragmentWith(SurfaceAccessibleI::Element(m_element),
				fragmentAttribute,inputGroup);
		fragmentCount=spOutputElements->fragmentCount();
	}

	if(!access(m_spDriverAccessible,"Driver Surface")) return;

	std::vector< sp<SurfaceAccessorI> > faces;
	std::vector< sp<SurfaceAccessorI> > barys;
	const I32 bindingsAccessed=accessBindings(m_spBindingAccessible,
			faces,barys,catalog<I32>("Bindings"),m_bindPrimitives,TRUE);
	if(bindingsAccessed<1)
	{
		return;
	}

	const String locatorPrefix=catalog<bool>("generateLocators")?
			catalog<String>("locatorPrefix"): "";

	const String signature=generateSignature(m_spDriverAccessible);
	storeSignature(m_spOutputAccessible,signature);

	storeParameters(m_spOutputAccessible,driverGroup,fragmentAttribute,
			inputGroup,bindingsAccessed,m_bindPrimitives,bindFragments,
			covert? covertKey: "",locatorPrefix,subdivDepth);

	const U32 elementCount=spOutputElements->count();

	sp<SurfaceAccessorI> spDriverVertices;
	if(m_conform)
	{
		if(!access(spDriverVertices,m_spDriverAccessible,
				e_primitive,e_vertices)) return;

		//* default -1 = no master
		m_masterBinding.clear();
		m_masterBinding.resize(elementCount,-1);

		const U32 driverPrimitiveCount=spDriverVertices->count();
		m_driverFacePrimitive.resize(driverPrimitiveCount*2,-1);

		sp<SurfaceAccessorI> spOutputProperties;
		if(!access(spOutputProperties,m_spOutputAccessible,
				e_primitive,e_properties)) return;

		const U32 primitiveCount=spOutputVertices->count();
		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
		{
			if(!spOutputProperties->integer(primitiveIndex,e_openCurve))
			{
				if(!warnedConformTriangle)
				{
					catalog<String>("warning")+="Conforming only supports"
							" open curves in Input Surface;";
					warnedConformTriangle=TRUE;
				}
				continue;
			}

			const U32 subCount=spOutputVertices->subCount(primitiveIndex);
			if(subCount<2)
			{
				continue;
			}

			const I32 masterIndex=spOutputVertices->integer(primitiveIndex,0);

			for(U32 subIndex=1;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,subIndex);
				m_masterBinding[pointIndex]=masterIndex;
			}
		}
	}

	//* TODO fragment-aware ranges
	//* NOTE conform unsafe by using m_driverFacePrimitive table
	//* NOTE covert access to Catalog is not thread safe
	//* NOTE partition filter setting is per surface
	if(m_fragmented || m_conform || covert || m_partitionDriver)
	{
		//* serial

		sp<SurfaceRange> spSerialRange=sp<SurfaceRange>(new SurfaceRange());

		if(m_fragmented)
		{
			spSerialRange->nonAtomic().add(0,fragmentCount-1);
		}
		else if(inputGroup.empty())
		{
			spSerialRange->nonAtomic().add(0,elementCount-1);
		}
		else
		{
			//* input group, not fragmented

			const Element groupElement=(m_element==e_primitive)?
					e_primitiveGroup: e_pointGroup;

			sp<SurfaceAccessorI> spElementGroup;
			access(spElementGroup,"Input Surface",groupElement,inputGroup);

			spSerialRange->nonAtomic().appendGroupSpans(spElementGroup);
		}

		run(-1,spSerialRange);
	}
	else
	{
		setAtomicRange(m_spOutputAccessible,
				m_bindPrimitives? e_pointsOfPrimitives: e_pointsOnly,
				inputGroup);

		OperatorThreaded::handle(a_rSignal);
	}

#if FE_SBO_DEBUG
	feLog("SurfaceBindOp::handle done\n");
#endif
}

sp<SurfaceI> SurfaceBindOp::subdivide(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String a_driverGroup,I32 a_depth,I32 a_flatDepth,Record& a_rSignal)
{
	sp<SurfaceI> spResultSurface;

	sp<Catalog> spOtherCatalog=m_spSubdivideOp;
	if(spOtherCatalog.isValid())
	{
		sp<SurfaceAccessibleI> spSubdividedAccessible=
				registry()->create("SurfaceAccessibleI");

		sp<Scope> spScope=registry()->create("Scope");
		spSubdividedAccessible->bind(spScope);

		spOtherCatalog->catalog<I32>("depth")=a_depth;
		spOtherCatalog->catalog<I32>("flatDepth")=a_flatDepth;
		spOtherCatalog->catalog< sp<Component> >("Input Surface")=
				a_rspAccessibleI;
		spOtherCatalog->catalog< sp<Component> >("Output Surface")=
				spSubdividedAccessible;

		sp<Component> spPriorOutput=surfaceOutput(a_rSignal);

		setSurfaceOutput(a_rSignal,spSubdividedAccessible);

		m_spSubdivideOp->handle(a_rSignal);

		setSurfaceOutput(a_rSignal,spPriorOutput);

		sp<SurfaceAccessorI> spSubdividedPoints;
		if(access(spSubdividedPoints,spSubdividedAccessible,
				e_point,e_position))
		{
//			spSubdividedAccessible->save("bind.rg");

			spResultSurface=spSubdividedAccessible->surface(a_driverGroup);
		}
	}

	return spResultSurface;
}

void SurfaceBindOp::run(I32 a_id,sp<SurfaceRange> a_spRange)
{
	const String fragmentAttribute=catalog<String>("FragmentAttr");
	const String inputGroup=catalog<String>("InputGroup");
	const String inputPartitionAttr=catalog<String>("InputPartitionAttr");
	const String pivotPoint=catalog<String>("PivotPoint");
	const BWORD limitDistance=catalog<bool>("LimitDistance");
	const Real maxDistance=limitDistance? catalog<Real>("MaxDistance"): -1.0;
	const BWORD scalable=catalog<bool>("Scalable");
	const BWORD generateLocators=
			(m_fragmented && catalog<bool>("generateLocators"));
	const String locatorPrefix=catalog<String>("locatorPrefix");

	std::vector< sp<SurfaceAccessorI> > faces;
	std::vector< sp<SurfaceAccessorI> > barys;
	const I32 bindingsAccessed=accessBindings(m_spBindingAccessible,
			faces,barys,catalog<I32>("Bindings"),m_bindPrimitives,FALSE);
	if(bindingsAccessed<1)
	{
		return;
	}

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	//* TODO don't redo fragmentation
	if(m_fragmented)
	{
		spOutputElements->fragmentWith(SurfaceAccessibleI::Element(m_element),
				fragmentAttribute,inputGroup);
	}

	sp<SurfaceAccessorI> spInputPartition;
	if(m_partitionDriver)
	{
		if(!access(spInputPartition,"Input Surface",
				m_element,inputPartitionAttr,e_error)) return;
	}

	sp<SurfaceAccessorI> spDriverVertices;
	sp<SurfaceAccessorI> spDriverProperties;
	if(m_conform)
	{
		if(!access(spDriverVertices,m_spDriverAccessible,
				e_primitive,e_vertices)) return;
		if(!access(spDriverProperties,m_spDriverAccessible,
				e_primitive,e_properties)) return;
	}

	I32 driverFace= -1;
	Barycenter<Real> driverBary;

	I32 deformedFace= -1;
	Barycenter<Real> deformedBary;

	sp<SurfaceAccessorI::FilterI> spFilter;

	for(SurfaceRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 index=it.value();

		String fragment;
		String locatorPreface;
		I32 filterCount=1;
		if(m_fragmented)
		{
			fragment=spOutputElements->fragment(index);
			if(!spOutputElements->filterWith(fragment,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}

			filterCount=spFilter->filterCount();
			if(!filterCount)
			{
				continue;
			}

			locatorPreface=locatorPrefix+"_"+fragment+"_";
		}

		if(m_partitionDriver)
		{
			const I32 reIndex=m_fragmented? spFilter->filter(0): index;

			const String partition=spInputPartition->string(reIndex);
			m_spDriver->setPartitionFilter(partition);

#if FE_SBO_VERBOSE
			feLog("SurfaceBindOp::handle element %d filter \"%s\"\n",
					index,partition.c_str());
#endif
		}

		SpatialVector pivot;

		if(pivotPoint=="Nearest")
		{
			set(pivot);
			Real bestDistance= -1.0;
			for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
			{
				const I32 reIndex=m_fragmented?
						spFilter->filter(filterIndex): index;
				const U32 subCount=spOutputElements->subCount(reIndex);

				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const SpatialVector point=
							spOutputElements->spatialVector(reIndex,subIndex);
					sp<SurfaceI::ImpactI> spImpact=
							m_spDriver->nearestPoint(point);
					const Real distance=spImpact->distance();
					if(bestDistance<0.0 || bestDistance>distance)
					{
						bestDistance=distance;
						pivot=point;
					}
				}
			}
		}
		else if(pivotPoint=="Center")
		{
			U32 sum=0;
			set(pivot);
			for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
			{
				const I32 reIndex=m_fragmented?
						spFilter->filter(filterIndex): index;
				const U32 subCount=spOutputElements->subCount(reIndex);

				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					pivot+=spOutputElements->spatialVector(reIndex,subIndex);
				}
				sum+=subCount;
			}
			if(!sum)
			{
				continue;
			}
			pivot*=1.0/Real(sum);
		}
		else
		{
			const I32 reIndex=m_fragmented? spFilter->filter(0): index;
			if(!spOutputElements->subCount(reIndex))
			{
				continue;
			}
			pivot=spOutputElements->spatialVector(reIndex,0);
		}

		if(m_conform)
		{
			const I32 reIndex=m_fragmented? spFilter->filter(0): index;
			const I32 masterIndex=m_masterBinding[reIndex];
			if(masterIndex>=0)
			{
				for(U32 binding=0;binding<bindingsAccessed;binding++)
				{
					driverFace=faces[binding]->integer(masterIndex);
					faces[binding]->set(reIndex,driverFace);

					if(driverFace<0)
					{
						continue;
					}

					const I32 driverPrimitiveIndex=
							m_driverFacePrimitive[driverFace];

					if(driverPrimitiveIndex>=0 && spDriverProperties->integer(
							driverPrimitiveIndex,e_openCurve))
					{
						const U32 subCount=
								spDriverVertices->subCount(driverFace);

						set(driverBary);
						driverBary[0]=subCount? nearestPointAlongCurve(pivot,
								spDriverVertices,
								driverFace)/(subCount-1.0): 0.0;
						driverBary[1]=1.0-driverBary[0];
					}
					else
					{
						driverBary=barys[binding]->spatialVector(
								masterIndex);
					}

					barys[binding]->set(reIndex,driverBary);
				}

				continue;
			}
		}

		std::vector< sp<SurfaceI::ImpactI> > impacts=
				m_spDriver->nearestPoints(pivot,maxDistance,bindingsAccessed);

		for(U32 binding=0;binding<bindingsAccessed;binding++)
		{
			const U32 impactCount=impacts.size();

			I32 face= -1;
			SpatialBary bary;
			set(bary);

			if(binding<impactCount && impacts[binding].isValid())
			{
				sp<SurfaceI::ImpactI>& rspImpact=impacts[binding];
				const Real distance=rspImpact->distance();
				if(!limitDistance || distance<=maxDistance)
				{
					face=rspImpact->triangleIndex();

					sp<SurfaceTriangles::Impact> spTriImpact=rspImpact;
					if(scalable && spTriImpact.isValid())
					{
						const SpatialVector normal=rspImpact->normal();
						const Real apart=
								dot(normal,pivot-rspImpact->intersection());
						const SpatialVector inPlane=pivot-normal*apart;

						bary.solve(spTriImpact->vertex0(),
								spTriImpact->vertex1(),
								spTriImpact->vertex2(),
								inPlane);
					}
					else
					{
						bary=rspImpact->barycenter();
					}

					if(generateLocators)
					{
						const SpatialVector normal=rspImpact->normal();
						const SpatialVector tangent=
								unitSafe(rspImpact->du()+rspImpact->dv());

						sp<SurfaceAccessorI> spLocPosition;
						if(!access(spLocPosition,m_spOutputAccessible,
								e_detail,locatorPreface+"P")) return;

						sp<SurfaceAccessorI> spLocNormal;
						if(!access(spLocNormal,m_spOutputAccessible,
								e_detail,locatorPreface+"N")) return;

						sp<SurfaceAccessorI> spLocTangent;
						if(!access(spLocTangent,m_spOutputAccessible,
								e_detail,locatorPreface+"T")) return;

						spLocPosition->set(0,pivot);
						spLocNormal->set(0,normal);
						spLocTangent->set(0,tangent);

					}

					if(m_conform)
					{
						FEASSERT(face<m_driverFacePrimitive.size());
						m_driverFacePrimitive[face]=
								rspImpact->primitiveIndex();
					}
				}
			}

			if(m_fragmented)
			{
				for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
				{
					const I32 filtered=spFilter->filter(filterIndex);

					faces[binding]->set(filtered,face);
					barys[binding]->set(filtered,bary);
				}
			}
			else
			{
				faces[binding]->set(index,face);
				barys[binding]->set(index,bary);
			}
#if FE_SBO_VERBOSE
			feLog("SurfaceBindOp::handle element %d bind %d/%d"
					" impactCount %d pivot %s face %d bary %s\n",
					index,binding,bindingsAccessed,impactCount,
					c_print(pivot),face,c_print(bary));
#endif
		}
	}
}

String SurfaceBindOp::retrieveSignature(
	sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String signatureName;
	signatureName.sPrintf("%sSignature",bindPrefix.c_str());

	sp<SurfaceAccessorI> spSignatureAccessor;
	if(!access(spSignatureAccessor,a_rspAccessibleI,
			e_detail,signatureName,e_warning))
	{
		return "";
	}

	return spSignatureAccessor->string(0);
}

void SurfaceBindOp::storeSignature(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String a_signature)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String signatureName;
	signatureName.sPrintf("%sSignature",bindPrefix.c_str());

	sp<SurfaceAccessorI> spSignatureAccessor;
	if(!access(spSignatureAccessor,a_rspAccessibleI,
			e_detail,signatureName,e_error,e_createMissing)) return;

	spSignatureAccessor->set(0,a_signature);

}

void SurfaceBindOp::discardSignature(sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String signatureName;
	signatureName.sPrintf("%sSignature",bindPrefix.c_str());

	discard(a_rspAccessibleI,e_detail,signatureName);
}

void SurfaceBindOp::retrieveParameters(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String& a_rDriverGroup,String& a_rFragmentAttribute,String& a_rInputGroup,
	I32& a_rBindings,BWORD& a_rBindPrimitives,BWORD& a_rBindFragments,
	String& a_rCovertKey,String& a_rLocators,I32& a_rSubdiv)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String attributeName;
	sp<SurfaceAccessorI> spAccessor;

	attributeName.sPrintf("%sDriverGroup",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rDriverGroup=spAccessor->string(0);
	}

	attributeName.sPrintf("%sFragment",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rFragmentAttribute=spAccessor->string(0);
	}

	attributeName.sPrintf("%sInputGroup",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rInputGroup=spAccessor->string(0);
	}

	attributeName.sPrintf("%sCount",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rBindings=spAccessor->integer(0);
	}

	attributeName.sPrintf("%sPrimitives",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rBindPrimitives=spAccessor->integer(0);
	}

	attributeName.sPrintf("%sFragments",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rBindFragments=spAccessor->integer(0);
	}

	attributeName.sPrintf("%sCovert",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rCovertKey=spAccessor->string(0);
	}

	attributeName.sPrintf("%sLocators",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rLocators=spAccessor->string(0);
	}

	attributeName.sPrintf("%sSubdiv",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rSubdiv=spAccessor->integer(0);
	}
}

void SurfaceBindOp::storeParameters(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String a_driverGroup,String a_fragmentAttribute,String a_inputGroup,
	I32 a_bindCount,BWORD a_bindPrimitives,BWORD a_bindFragments,
	String a_covertKey,String a_locatorPrefix,I32 a_subdiv)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String attributeName;
	sp<SurfaceAccessorI> spAccessor;

	attributeName.sPrintf("%sDriverGroup",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_driverGroup);
	}

	attributeName.sPrintf("%sFragment",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_fragmentAttribute);
	}

	attributeName.sPrintf("%sInputGroup",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_inputGroup);
	}

	attributeName.sPrintf("%sCount",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_bindCount);
	}

	attributeName.sPrintf("%sPrimitives",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_bindPrimitives);
	}

	attributeName.sPrintf("%sFragments",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_bindFragments);
	}

	attributeName.sPrintf("%sCovert",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_covertKey);
	}

	attributeName.sPrintf("%sLocators",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_locatorPrefix);
	}

	attributeName.sPrintf("%sSubdiv",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_subdiv);
	}
}

void SurfaceBindOp::discardParameters(sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String attributeName;
	sp<SurfaceAccessorI> spAccessor;

	attributeName.sPrintf("%sDriverGroup",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sFragment",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sInputGroup",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sCount",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sPrimitives",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sFragments",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sCovert",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sLocators",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sSubdiv",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);
}

I32 SurfaceBindOp::accessBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	std::vector< sp<SurfaceAccessorI> >& a_rFaces,
	std::vector< sp<SurfaceAccessorI> >& a_rBarys,
	I32 a_bindCount,BWORD a_bindPrimitives,BWORD a_allowCreation)
{
	const SurfaceAccessibleI::Element element=a_bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;

	const SurfaceAccessibleI::Creation create=a_allowCreation?
			SurfaceAccessibleI::e_createMissing:
			SurfaceAccessibleI::e_refuseMissing;

	U32 bindingsAccessed=0;

	const String bindPrefix=catalog<String>("BindPrefix");

	a_rFaces.resize(a_bindCount);
	a_rBarys.resize(a_bindCount);

	for(U32 binding=0;binding<a_bindCount;binding++)
	{
		String faceName;
		String baryName;
		faceName.sPrintf("%sFace%d",bindPrefix.c_str(),binding);
		baryName.sPrintf("%sBary%d",bindPrefix.c_str(),binding);

		a_rFaces[binding]=a_rspAccessibleI->accessor(element,faceName,create);
		if(a_rFaces[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp face attribute"
						" \"%s\" not accessible\n",faceName.c_str());
				catalog<String>("error")=
						"inaccessible face attribute \""+faceName+"\"";
				return -1;
			}
			break;
		}
		else if(a_allowCreation)
		{
#if FALSE
			//* force immediate creation
			a_rFaces[binding]->set(0,0);
#else
			const U32 elementCount=a_rFaces[binding]->count();
			for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
			{
				a_rFaces[binding]->set(elementIndex,-1);
			}
#endif
		}

		a_rBarys[binding]=
				a_rspAccessibleI->accessor(element,baryName,create);
		if(a_rBarys[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp barycentric attribute"
						" \"%s\" not accessible\n",baryName.c_str());
				catalog<String>("error")="inaccessible barycentric"
						" attribute \""+baryName+"\"";
				return -1;
			}
			break;
		}
		else if(a_allowCreation)
		{
			//* force immediate creation
			a_rBarys[binding]->set(0,Barycenter<Real>(0.0,0.0));
		}

		bindingsAccessed++;
	}

	return bindingsAccessed;
}

void SurfaceBindOp::discardBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	I32 a_bindCount,BWORD a_bindPrimitives)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	const Element element=a_bindPrimitives? e_primitive: e_point;

	for(U32 binding=0;binding<a_bindCount;binding++)
	{
		String faceName;
		String baryName;
		faceName.sPrintf("%sFace%d",bindPrefix.c_str(),binding);
		baryName.sPrintf("%sBary%d",bindPrefix.c_str(),binding);

		discard(a_rspAccessibleI,element,faceName);
		discard(a_rspAccessibleI,element,baryName);
	}
}
