/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_SurfaceBindOp_h__
#define __operator_SurfaceBindOp_h__

namespace fe
{

/**************************************************************************//**
	@brief Operator to correlate a attachable surface with a driving surface

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT SurfaceBindOp:
	public OperatorThreaded,
	public Initialize<SurfaceBindOp>
{
	public:

						SurfaceBindOp(void)									{}
virtual					~SurfaceBindOp(void)								{}

		void			initialize(void);

						//* As BindI
virtual	void			bind(sp<SignalerI> a_spSignalerI,
								sp<Layout> a_spLayout);
						//* As HandlerI
virtual	void			handle(Record& a_rSignal);

virtual	void			run(I32 a_id,sp<SurfaceRange> a_spRange);

	protected:
		sp<SurfaceI>	subdivide(sp<SurfaceAccessibleI>& a_rspAccessibleI,
								String a_driverGroup,I32 a_depth,
								I32 a_flatDepth,Record& a_rSignal);

		String			retrieveSignature(
								sp<SurfaceAccessibleI>& a_rspAccessibleI);
		void			storeSignature(sp<SurfaceAccessibleI>& a_rspAccessibleI,
								String a_signature);
		void			discardSignature(
								sp<SurfaceAccessibleI>& a_rspAccessibleI);

		void			retrieveParameters(
								sp<SurfaceAccessibleI>& a_rspAccessibleI,
								String& a_rDriverGroup,
								String& a_rFragmentAttribute,
								String& a_rInputGroup,
								I32& a_rBindings,BWORD& a_rBindPrimitives,
								BWORD& a_rBindFragments,String& a_rCovertKey,
								String& a_rLocators,I32& a_rSubdiv);
		void			storeParameters(
								sp<SurfaceAccessibleI>& a_rspAccessibleI,
								String a_driverGroup,String a_fragmentAttribute,
								String a_inputGroup,
								I32 a_bindCount,BWORD a_bindPrimitives,
								BWORD a_bindFragments,String a_covertKey,
								String a_locators,I32 a_subdiv);
		void			discardParameters(
								sp<SurfaceAccessibleI>& a_rspAccessibleI);

		I32				accessBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
								std::vector< sp<SurfaceAccessorI> >& a_rFaces,
								std::vector< sp<SurfaceAccessorI> >& a_rBarys,
								I32 a_bindCount,
								BWORD a_bindPrimitives,
								BWORD a_allowCreation);
		void			discardBindings(
								sp<SurfaceAccessibleI>& a_rspAccessibleI,
								I32 a_bindCount,BWORD a_bindPrimitives);

		sp<SurfaceAccessibleI>					m_spBindingAccessible;
		sp<SurfaceAccessibleI>					m_spOutputAccessible;
		sp<SurfaceAccessibleI>					m_spDriverAccessible;

		sp<SurfaceI>							m_spDriver;

		sp<OperatorSurfaceI>					m_spSubdivideOp;

		BWORD									m_fragmented;
		BWORD									m_bindPrimitives;
		BWORD									m_conform;
		BWORD									m_partitionDriver;

		Element									m_element;
		Attribute								m_attribute;

		std::vector<I32>						m_masterBinding;
		std::vector<I32>						m_driverFacePrimitive;
};

} /* namespace */

#endif /* __operator_SurfaceBindOp_h__ */
