/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

namespace fe
{

FE_DL_PUBLIC ThreadDefaultInitFunction* Thread::ms_fnDefaultInit=NULL;
FE_DL_PUBLIC ThreadInitFunction* Thread::ms_fnInit=NULL;
FE_DL_PUBLIC ThreadInterruptFunction* Thread::ms_fnInterrupt=NULL;
FE_DL_PUBLIC ThreadJoinFunction* Thread::ms_fnJoin=NULL;
FE_DL_PUBLIC ThreadFinishFunction* Thread::ms_fnFinish=NULL;
FE_DL_PUBLIC ThreadInterruptionFunction* Thread::ms_fnInterruption=NULL;
FE_DL_PUBLIC ThreadConcurrencyFunction* Thread::ms_fnConcurrency=NULL;

FE_DL_PUBLIC ThreadGroupInitFunction* Thread::ms_fnGroupInit=NULL;
FE_DL_PUBLIC ThreadGroupCreateFunction* Thread::ms_fnGroupCreate=NULL;
FE_DL_PUBLIC ThreadGroupJoinAllFunction* Thread::ms_fnGroupJoinAll=NULL;
FE_DL_PUBLIC ThreadGroupFinishFunction* Thread::ms_fnGroupFinish=NULL;

FE_DL_PUBLIC BWORD Thread::ms_checked=FALSE;

Thread::Thread(void):
	m_pRawThread(NULL)
{
	confirm();

	if(ms_fnDefaultInit)
	{
		m_pRawThread=ms_fnDefaultInit();
	}

	FEASSERT(m_pRawThread);
}

Thread::Thread(Functor* a_pFunctor):
	m_pRawThread(NULL)
{
	confirm();

	if(ms_fnInit)
	{
		m_pRawThread=ms_fnInit(a_pFunctor);
	}

	FEASSERT(m_pRawThread);
}

Thread::Thread(void* a_pRawThread):
	m_pRawThread(a_pRawThread)
{
	FEASSERT(m_pRawThread);
}

Thread::~Thread(void)
{
	if(m_pRawThread)
	{
		if(ms_fnFinish)
		{
			ms_fnFinish(m_pRawThread);
		}
		m_pRawThread=NULL;
	}
}

BWORD Thread::interrupt(void)
{
	if(!ms_fnInterrupt)
	{
		return FALSE;
	}

	ms_fnInterrupt(m_pRawThread);
	return TRUE;
}

BWORD Thread::join(void)
{
	if(!ms_fnJoin)
	{
		return FALSE;
	}

	ms_fnJoin(m_pRawThread);
	return TRUE;
}

//* static
BWORD Thread::interruptionPoint(void)
{
	if(!ms_fnInterruption)
	{
		return FALSE;
	}

	ms_fnInterruption();
	return TRUE;
}

//* static
I32 Thread::hardwareConcurrency(void)
{
	confirm();

	if(!ms_fnConcurrency)
	{
		return 1;
	}

	return ms_fnConcurrency();
}


Thread::Functor::Functor(void):
	m_pOriginal(this)
{
//	printf("Thread::Functor::Functor %p\n",this);
}

Thread::Functor::~Functor(void)
{
//	printf("Thread::Functor::~Functor %p\n",this);
}

void Thread::Functor::operator()(void)
{
//	printf("Thread::Functor::operator() %p %p\n",this,m_pOriginal);

	m_pOriginal->operate();
}

void Thread::Functor::operate(void)
{
	printf("Thread::Functor::operate doing nothing as base class\n");
}

Thread::Group::Group(void):
	m_pRawThreadGroup(NULL)
{
	confirm();

	if(Thread::ms_fnGroupInit)
	{
		m_pRawThreadGroup=Thread::ms_fnGroupInit();
	}
}

Thread::Group::~Group(void)
{
	if(Thread::ms_fnGroupFinish)
	{
		Thread::ms_fnGroupFinish(m_pRawThreadGroup);
	}
}

Thread* Thread::Group::createThread(Functor* a_pFunctor)
{
	if(Thread::ms_fnGroupCreate)
	{
		void* pRawThread=Thread::ms_fnGroupCreate(m_pRawThreadGroup,a_pFunctor);
		return new Thread(pRawThread);
	}
	return NULL;
}

BWORD Thread::Group::joinAll(void)
{
	if(Thread::ms_fnGroupJoinAll)
	{
		Thread::ms_fnGroupJoinAll(m_pRawThreadGroup);
		return TRUE;
	}
	return FALSE;
}

//* static
BWORD Thread::confirm(void)
{
	if(ms_fnInit)
	{
		return TRUE;
	}
	if(ms_checked)
	{
		return FALSE;
	}
	ms_checked=TRUE;

	const String verbose=System::getVerbose();

	DL_Loader *pLoader=new DL_Loader();
	FEASSERT(pLoader);

	BWORD success=FALSE;
	const BWORD adaptname=TRUE;

	String nameOrder=FE_STRING(FE_THREAD_ORDER);
	String name;
	while(!success && !(name=nameOrder.parse("\"",":")).empty())
	{
		success=pLoader->openDL(name,adaptname);
	}

	if(!success)
	{
		if(verbose!="none")
		{
			printf("  no thread support\n");
		}
		delete pLoader;
		return FALSE;
	}

	bool (*thread_init)(void);
	thread_init=(bool (*)(void))pLoader->findDLFunction("thread_init");

	if(!thread_init)
	{
		if(verbose!="none")
		{
			fprintf(stderr,"  thread_init not found\n");
		}
		delete pLoader;
		return FALSE;
	}

	if(verbose=="all")
	{
		printf("  using %s (for threading)\n",name.c_str());
	}

	return thread_init();

	//* TODO retain, close, and delete pLoader later
}

} /* namespace */

